import pytest

try:
    from django.conf import settings
except ImportError:
    pytest.skip()

from django.http import HttpResponse, HttpResponseRedirect
from django.test import RequestFactory

from unpoly.contrib.django import UnpolyMiddleware

settings.configure()


def get_response(request):
    return HttpResponse("")


@pytest.fixture
def mw():
    return UnpolyMiddleware(get_response)


@pytest.fixture
def factory():
    return RequestFactory()


def test_django_middleware_headers(mw, factory):
    req = factory.get("/test", HTTP_X_UP_VERSION="2.2.1")
    response = mw(req)
    assert response.headers["X-Up-Method"] == "GET"
    assert "X-Up-Location" not in response.headers


def test_django_up_cookie(mw, factory):
    # Setting the cookie
    req = factory.post("/test")
    response = mw(req)
    assert response.cookies["_up_method"].value == "POST"

    # Deleting the cookie if set
    req = factory.get("/test", HTTP_COOKIE="_up_method=POST", HTTP_X_UP_VERSION="2.2.1")
    response = mw(req)
    assert response.cookies["_up_method"].value == ""

    # Not doing anything
    req = factory.get("/test", HTTP_X_UP_VERSION="2.2.1")
    response = mw(req)
    assert "_up_method" not in response.cookies


def test_django_redirect_up_params(factory):
    def get_response(request):
        request.up.set_title("Testing")
        return HttpResponseRedirect("/redirect?test=abc")

    mw = UnpolyMiddleware(get_response)

    req = factory.get("/test", HTTP_X_UP_VERSION="2.2.1")
    response = mw(req)
    assert response.headers["Location"] == "/redirect?test=abc&_up_title=Testing"



import json
from datetime import datetime, timezone

from unpoly.adapter import BaseAdapter
from unpoly.options import Options


def test_parsing_from_request_headers():
    options = Options.parse(
        {
            "version": "2.2.1",
            "target": "[up-main~=modal]",
            "mode": "modal",
            "context": '{"test": "test"}',
            "fail_target": "[up-main~=fail]",
            "fail_mode": "fail",
            "fail_context": '{"fail": "fail"}',
            "validate": "field",
            "reload_from_time": 1608730818,
            "dismiss_layer": "null",
        },
        BaseAdapter(),
    )

    assert options.version == "2.2.1"
    assert options.target == "[up-main~=modal]"
    assert options.mode == "modal"
    assert options.context == {"test": "test"}
    assert options.fail_target == "[up-main~=fail]"
    assert options.fail_mode == "fail"
    assert options.fail_context == {"fail": "fail"}
    assert options.validate == "field"
    assert options.reload_from_time == datetime(
        2020, 12, 23, 13, 40, 18, tzinfo=timezone.utc
    )
    assert options.dismiss_layer == {}


def test_parse_broken_time():
    options = Options.parse({"reload_from_time": "no_date_time"}, BaseAdapter())
    assert options.reload_from_time == datetime(1970, 1, 1, 0, 0, tzinfo=timezone.utc)


def test_parse_broken_context():
    options = Options.parse({"context": "'a'a"}, BaseAdapter())
    assert options.context == {}


def test_parsing_with_diff():
    initial = {"existing": 1, "to_delete": True, "do_no_touch": "yes"}
    diff = {"to_delete": None, "existing": 2, "new_thing": 3}
    options = Options.parse(
        {"context": json.dumps(initial), "context_diff": json.dumps(diff)},
        BaseAdapter(),
    )
    assert options.initial_context == initial
    assert options.context_diff == diff
    assert options.context == {"existing": 2, "do_no_touch": "yes", "new_thing": 3}


def test_context_diff():
    options = Options(context={"existing": 1, "to_delete": True, "do_no_touch": "yes"})
    del options.context["to_delete"]
    options.context["existing"] = 2
    options.context["new_thing"] = 3
    assert options.context_diff == {
        "to_delete": None,
        "existing": 2,
        "new_thing": 3,
    }


def test_serialize():
    assert Options(target="target").serialize(BaseAdapter()) == {}

    options = Options(
        context={"test": "test"},
        events=[{"test": "event"}],
        title="test",
        accept_layer={},
        dismiss_layer={"test": "layer"},
    )
    options.context["change"] = "yes"  # Triggers diff rendering
    options.server_target = "new_target"  # Trigger target

    assert options.serialize(BaseAdapter()) == {
        "context": '{"change":"yes"}',  # This is just the diff!
        "events": '[{"test":"event"}]',
        "title": "test",
        "accept_layer": "null",
        "dismiss_layer": '{"test":"layer"}',
        "target": "new_target",
    }



from datetime import datetime, timezone
from unittest.mock import Mock

from unpoly.adapter import SimpleAdapter
from unpoly.up import Unpoly


def make_up(*args, **kwargs):
    adapter = SimpleAdapter(*args, **kwargs)
    up = Unpoly(adapter)
    return up, adapter


def test_is_up_request():
    up, _ = make_up()
    assert bool(up) is not True

    up, _ = make_up(headers={"X-Up-Version": "2.2.1"})
    assert bool(up) is True


def test_needs_cookie():
    up, _ = make_up()
    assert up.needs_cookie is False

    up, _ = make_up(method="POST", headers={"X-Up-Version": "2.2.1"})
    assert up.needs_cookie is False

    up, _ = make_up(method="POST")
    assert up.needs_cookie is True


def test_set_title():
    up, _ = make_up()
    up.set_title("Testing")
    assert up.options.title == '"Testing"'


def test_set_title_v2():
    up, _ = make_up(headers={"X-Up-Version": "2.2.1"})
    up.set_title("Testing")
    assert up.options.title == "Testing"


def test_emit_events():
    up, _ = make_up()

    up.emit("test", {"k": "v"})
    up.layer.emit("test2", {"k2": "v2"})

    assert up.options.events == [
        {"type": "test", "k": "v"},
        {"type": "test2", "layer": "current", "k2": "v2"},
    ]


def test_layer():
    up, _ = make_up(
        headers={
            "X-Up-Mode": "modal",
            "X-Up-Fail-Mode": "fail",
            "X-Up-Context": '{"k":"v"}',
            "X-Up-Fail-Context": '{"k2":"v2"}',
        }
    )
    assert up.layer.mode == "modal"
    assert up.layer.context == {"k": "v"}
    assert up.fail_layer.mode == "fail"
    assert up.fail_layer.context == {"k2": "v2"}


def test_layer_accept():
    up, _ = make_up(headers={"X-Up-Mode": "modal"})

    assert up.layer.is_overlay
    up.layer.accept("test")
    assert up.options.accept_layer == "test"

    up.layer.accept()
    assert up.options.accept_layer == {}  # Ugly, but needed for roundtripping


def test_layer_dismiss():
    up, _ = make_up(headers={"X-Up-Mode": "modal"})

    assert up.layer.is_overlay
    up.layer.dismiss("test")
    assert up.options.dismiss_layer == "test"


def test_layer_mode():
    up, _ = make_up(headers={"X-Up-Mode": "modal"})
    assert up.layer.is_overlay

    up, _ = make_up(headers={"X-Up-Mode": "root"})
    assert up.layer.is_root


def test_cache():
    up, _ = make_up(headers={"X-Up-Mode": "modal"})

    up.cache.clear("test/*")
    assert up.options.clear_cache == "test/*"

    up.cache.keep()
    assert up.options.clear_cache == "false"


def test_parsing_from_request():
    up, _ = make_up(
        headers={
            "X-Up-Version": "2.2.1",
            "X-Up-Target": "[up-main~=modal]",
            "X-Up-Mode": "modal",
            "X-Up-Context": '{"test":"test"}',
            "X-Up-Fail-Target": "[up-main~=fail]",
        },
        params={
            "_up_fail_mode": "fail",
            "_up_fail_context": '{"fail":"fail"}',
            "_up_validate": "field",
            "_up_reload_from_time": "1608730818",
        },
    )

    assert up.version == "2.2.1"
    assert up.target == "[up-main~=modal]"
    assert up.mode == "modal"
    assert up.context == {"test": "test"}
    assert up.fail_target == "[up-main~=fail]"
    assert up.fail_mode == "fail"
    assert up.fail_context == {"fail": "fail"}
    assert up.validate == "field"
    assert up.reload_from_time == datetime(
        2020, 12, 23, 13, 40, 18, tzinfo=timezone.utc
    )


def test_target_change():
    up, _ = make_up(headers={"X-Up-Target": "[up-main~=modal]"})
    assert not up.options.target_changed
    up.target = "new_target"
    assert up.options.target_changed
    assert up.options.server_target == "new_target"


def test_finalize_cookie():
    up, adapter = make_up(method="POST")
    adapter.set_cookie = Mock()
    up.finalize_response(None)
    adapter.set_cookie.assert_called_once_with(None, True)

    up, adapter = make_up()
    adapter.set_cookie = Mock()
    up.finalize_response(None)
    adapter.set_cookie.assert_called_once_with(None, False)


def test_no_rewrite_if_not_unpoly():
    up, adapter = make_up()
    adapter.set_headers = Mock()
    adapter.set_redirect_uri = Mock()
    up.finalize_response(None)
    adapter.set_headers.assert_not_called()
    adapter.set_redirect_uri.assert_not_called()


def test_response_headers():
    up, adapter = make_up(
        location="/test", method="POST", headers={"X-Up-Version": "2.1.1"}
    )
    up.finalize_response(None)
    assert adapter.response_headers == {"X-Up-Method": "POST"}


def test_response_headers_with_location():
    up, adapter = make_up(
        location="/test?_up_test=123", method="POST", headers={"X-Up-Version": "2.1.1"}
    )
    up.finalize_response(None)
    assert adapter.response_headers == {"X-Up-Location": "/test", "X-Up-Method": "POST"}


def test_redirect_params():
    up, adapter = make_up(headers={"X-Up-Version": "2.1.1"}, redirect_uri="/redirect")
    up.set_title("test")
    up.context["new"] = "yes"
    up.finalize_response(None)
    assert adapter.response_headers is None
    assert (
        adapter.response_redirect_uri
        == "/redirect?_up_title=test&_up_context_diff=%7B%22new%22%3A%22yes%22%7D"
    )


def test_up_param_stripping():
    up, adapter = make_up(
        location="/test?a=b&_up_title=test&_up_target=gone",
        headers={"X-Up-Version": "2.1.1"},
    )
    up.finalize_response(None)
    assert adapter.response_headers == {
        "X-Up-Location": "/test?a=b",  # _up_ params are dropped (history support)
        "X-Up-Method": "GET",
    }



import copy
from datetime import datetime, timezone
from functools import partial
from typing import Any, List, Mapping, Optional

import attr

RENDER_OPTS = {
    "accept_layer",
    "dismiss_layer",
    "clear_cache",
    "events",
    "title",
    "events",
}


def ts_to_datetime(data: str) -> Optional[datetime]:
    try:
        ts = int(data)
        return datetime.fromtimestamp(ts, timezone.utc)
    except (ValueError, TypeError):
        return datetime.fromtimestamp(0, timezone.utc)


@attr.define(kw_only=True)
class Options:
    # mostly request options
    version: Optional[str] = None
    target: Optional[str] = None
    mode: Optional[str] = None
    context: Optional[Mapping[str, Any]] = None
    fail_target: Optional[str] = None
    fail_mode: Optional[str] = None
    fail_context: Optional[Mapping[str, Any]] = None
    validate: Optional[str] = None
    reload_from_time: Optional[datetime] = None

    # internal params
    server_target: Optional[str] = attr.field(init=False, default=None)
    initial_context: Optional[Mapping[str, Any]] = attr.field(init=False)

    # explicit response options
    title: Optional[str] = None
    clear_cache: Optional[str] = None
    accept_layer: Optional[str] = None
    dismiss_layer: Optional[str] = None
    events: List[Mapping[str, Any]] = None

    def __attrs_post_init__(self):
        if self.context is None:
            self.context = {}
        self.initial_context = copy.deepcopy(self.context)

    @property
    def target_changed(self):
        return self.server_target and self.server_target != self.target

    @property
    def context_diff(self):
        old = self.initial_context
        new = self.context
        old_keys = set(old)
        new_keys = set(new)
        result = {}
        for key in old_keys - new_keys:
            result[key] = None
        for key in new_keys - old_keys:
            result[key] = new[key]
        for key in new_keys & old_keys:
            if old[key] != new[key]:
                result[key] = new[key]
        return result

    @classmethod
    def parse(cls, options, adapter):
        if "reload_from_time" in options:
            options["reload_from_time"] = ts_to_datetime(options["reload_from_time"])
        for opt in {
            "events",
            "context",
            "context_diff",
            "fail_context",
            "dismiss_layer",
            "accept_layer",
            "context_diff",
        }:
            if opt in options:
                options[opt] = adapter.deserialize_data(options[opt])
                if opt in {"accept_layer", "dismiss_layer"}:
                    options[opt] = {}
        context_diff = options.pop("context_diff", None)

        options = cls(**options)
        # Apply the passed context diff
        if context_diff:
            for k, v in context_diff.items():
                if v is None and k in options.context:
                    del options.context[k]
                else:
                    options.context[k] = v
        return options

    def _serialize(self, inst, field, value, *, adapter):
        if field.name in {"accept_layer", "dismiss_layer", "events"}:
            return "null" if value == {} else adapter.serialize_data(value)
        return value

    def serialize(self, adapter):
        serialized_options = attr.asdict(
            self,
            filter=lambda attr, value: attr.name in RENDER_OPTS and value is not None,
            value_serializer=partial(self._serialize, adapter=adapter),
        )

        # Update with new target, if it changed
        if self.target_changed:
            serialized_options["target"] = self.server_target

        # Update with context changes, see https://unpoly.com/X-Up-Context
        diff = self.context_diff
        if diff:
            serialized_options["context"] = adapter.serialize_data(diff)

        return serialized_options



"""
Adapters
========

This library uses adapters to implement the "I/O parts", namely reading and writing headers
from HTTP requests and response. To adapt this library to a new framework it is neccessary to
subclass :class:`.BaseAdapter` and then initialize :class:`unpoly.up.Unpoly` with it. Since
:class:`unpoly.up.Unpoly` is initialized once per request, it makes sense to pass the request
to the adapter via init, but there is no requirement for this -- the adapter could simply
access the request via thread-locals or other means.

As an example the usage with the builtin Django Adapter looks like this:

.. code-block:: python

    from unpoly import Unpoly
    from unpoly.contrib.django import DjangoAdapter
    # The following code is run in a middleware, once per request
    adapter = DjangoAdapter(request)
    # Attach `Unpoly` to the request, so views can easily acces it
    request.up = Unpoly(adapter)
    # Actually execute the view and get the response object
    response = handle_view(request)
    # Tell `Unpoly` to set the relevant `X-Up-*` headers
    request.up.finalize_reponse(response)

"""
import json
from typing import Mapping, Optional


class BaseAdapter:
    """
    Provides the entrypoint for other frameworks to use this library.

    Implements common functionality that is not often overriden as well
    as framework specific hooks.
    """

    def request_headers(self) -> Mapping[str, str]:
        """Reads the request headers from the current request.

        Needs to be implemented."""
        raise NotImplementedError  # pragma: no cover

    def request_params(self) -> Mapping[str, str]:
        """Reads the GET params from the current request.

        Needs to be implemented."""
        raise NotImplementedError  # pragma: no cover

    def redirect_uri(self, response) -> Optional[str]:
        """Returns the redirect target of a response or None if the response
        is not a redirection (ie if it's status code is not in the range 300-400).

        Needs to be implemented."""
        raise NotImplementedError  # pragma: no cover

    def set_redirect_uri(self, response, uri):
        """Set a new redirect target for the current response. This is used to
        pass unpoly parameters via GET params through redirects.

        Needs to be implemented."""
        raise NotImplementedError  # pragma: no cover

    def set_headers(self, response, headers):
        """Set headers like `X-Up-Location` on the current response.

        Needs to be implemented."""
        raise NotImplementedError  # pragma: no cover

    def set_cookie(self, response, needs_cookie=False):
        """Set or delete the `_up_method <https://unpoly.com/_up_method>`_ cookie.

        The implementation should set the cookie if `needs_cookie` is `True` and
        otherwise remove it if set.

        Needs to be implemented."""
        raise NotImplementedError  # pragma: no cover

    @property
    def method(self):
        """Exposes the current request's method (GET/POST etc)

        Needs to be implemented."""
        raise NotImplementedError  # pragma: no cover

    @property
    def location(self):
        """Exposes the current request's location (path including query params)

        Needs to be implemented."""
        raise NotImplementedError  # pragma: no cover

    def deserialize_data(self, data: str):
        """Deserializes data passed in by Unpoly.

        By default it simply reads it as JSON, but can be overriden if custom
        decoders are needed.
        """
        try:
            return json.loads(data)
        except json.JSONDecodeError:
            return None

    def serialize_data(self, data) -> str:
        """Serializes the data for passing it to Unpoly.

        By default it simply serializes it as JSON, but can be overriden if custom
        encoders are needed.
        """
        return json.dumps(data, separators=(",", ":"), ensure_ascii=True)


class SimpleAdapter(BaseAdapter):
    def __init__(
        self,
        method="GET",
        location="/",
        headers=None,
        params=None,
        redirect_uri=None,
    ):
        self._method = method
        self._location = location
        self.headers = headers or {}
        self.params = params or {}
        self._redirect_uri = redirect_uri
        self.response_redirect_uri = None
        self.response_headers = None
        self.cookie = None

    def request_headers(self) -> Mapping[str, str]:
        return self.headers

    def request_params(self) -> Mapping[str, str]:
        return self.params

    def redirect_uri(self, response) -> Optional[str]:
        return self._redirect_uri

    def set_redirect_uri(self, response, uri):
        self.response_redirect_uri = uri

    def set_headers(self, response, headers):
        self.response_headers = headers

    def set_cookie(self, response, needs_cookie):
        self.cookie = needs_cookie

    @property
    def method(self):
        return self._method

    @property
    def location(self):
        return self._location
